/********************************************************************************
  uniprime7_mrna_get_highlighyed_probes_flex.c
  Roderic Guigo Serra, 2003
  check uniquenss of primers
  modified by Sarah Djebali and Sylvain Foissac on Dec 2006
  takes as input:
  - a list of mrna race product sequences in tbl format and uppercase letters
  - a list of probes in tbl format and uppercase letters
  and reports all the probes highlighted, or more precisely all the alignments 
  of the probes against all the mrna race product sequences

Used for a huge list of probes (several millions) against a sublist of the transcriptome
does not make use of neither the parameters TERMLEN, PLENG, PIDTY, nor the parameters
GXSEQLENGTH, MAXNOGXSEQ, MAXNOPROBE


Be careful: 
----------
do not reverse complement the transcriptome, this should be done outside
********************************************************************************/

#include <stdio.h>
#include <time.h>

#define MAXSTRINGG 200
#define PROBELENGTH 50
#define MAXSTRINGP 200


/* ./uniprime7_mrna_get_highlighted_probes_flex raceprodseqfile.tbl probefile.tbl maxnoraceprod maxraceprodseqlg maxnoprobes maxnbmismatch */

void shortime(void)
{
  time_t mytime = time(0); /* not 'long' */
  fprintf(stderr,"time: %s ", ctime(&mytime));
  fflush(stderr);
}

main(int argc, char *argv[])
{
  FILE *file;
  char * * GxSeq;
  char * * nameGxSeq; 
  int * LengthGxSeq;
  char * * Probe;
  char * * nameProbe;
  int * LengthProbe;

  char subseq[PROBELENGTH];
  long i,j,l,m,n,p,q;
  int maxnogseq, gxseqlength, maxnoprobe;
  int maxnbmismatch, mismatch;

  /* usage */
  if (argc != 7)
    {
      fprintf(stderr,"UNIPRIME MRNA: check uniqueness of probe sequences in a set of mrna race products\n");
      fprintf(stderr,"WARNING: probes and mrna race product sequences must be UPPER-CASE\n");
      fprintf(stderr,"Usage:\n");
      fprintf(stderr,"./uniprime7_mrna_get_highlighyed_probes_flex mrnaseqfile.tbl probefile.tbl maxnomrna maxmrnaseqlg maxnoprobes maxnbmismatch\n");
      exit(1);
    }

  /***************************************************************************************************************
   read the mrna sequences and the command line arguments relative to them and compute the lengths of the sequences
   ***************************************************************************************************************/
  maxnogseq = atoi (argv[3]);
  gxseqlength = atoi (argv[4]);

  if ((file=fopen(argv[1], "r"))==NULL)
    {
      fprintf(stderr,"The mrna sequence file %s cannot be opened\n",argv[1]);
      exit(1);
    }

  /* allocate GxSeq, nameGxSeq and LengthGxSeq */
  if(!(GxSeq=(char * *) malloc(maxnogseq*sizeof(char *))))
    {
      fprintf(stderr,"uniprime: not enough memory\n");
      exit(1); 
    }

  if(!(nameGxSeq=(char * *) malloc(maxnogseq*sizeof(char *))))
    {
      fprintf(stderr,"uniprime: not enough memory\n");
      exit(1); 
    }

  if(!(LengthGxSeq=(int *) malloc(maxnogseq*sizeof(int))))
    {
      fprintf(stderr,"uniprime: not enough memory\n");
      exit(1); 
    }

  for(p=0; p<maxnogseq; p++)
    {
      if(!(GxSeq[p]= (char *) malloc(gxseqlength*sizeof(char))))
	{
	  fprintf(stderr,"uniprime: not enough memory\n");
	  exit(1);
	}
      if(!(nameGxSeq[p] =  (char *) malloc(MAXSTRINGG*sizeof(char))))
	{
	  fprintf(stderr,"uniprime: not enough memory\n");
	  exit(1);
	}
    }
  fprintf(stderr,"I have done the double allocation of GxSeq and nameGxSeq, and the simple allocation of LengthGxSeq\n");
  
  q=0;
  while (fscanf(file, "%s %s\n", nameGxSeq[q], GxSeq[q]) !=EOF)
    {
      //fprintf(stderr,"I am in the mrna sequence file scan, q is %d and nameGxSeq[q] is %s\n", q, nameGxSeq[q]);
      q++;
    }
  fprintf(stderr,"I have done the mrna sequence file scan and q is %d\n", q); 

  for (p=0;p<q;p++)
    {
      LengthGxSeq[p] = strlen(GxSeq[p]); 
      //fprintf(stderr,"The length of the mrna sequence no %d is %d\n", p, LengthGxSeq[p]);
    }
  fprintf(stderr,"I have computed the length of each mrna sequence\n"); 



   /***************************************************************************************************************
   read the probe sequences and the command line arguments relative to them and compute the lengths of the sequences
   ***************************************************************************************************************/
  maxnoprobe = atoi (argv[5]);
  
  if ((file=fopen(argv[2], "r"))==NULL)
    {
      fprintf(stderr,"The probe file %s cannot be opened\n",argv[2]);
      exit(1);
    }

  /* allocate Probe, nameProbe and LengthProbe */
  if(!(Probe = (char * *)  malloc(maxnoprobe * sizeof(char *)))) 
    {
      fprintf(stderr,"uniprime: not enough memory\n");
      exit(1);
    }

  if(!(nameProbe = (char * *)  malloc(maxnoprobe * sizeof(char *)))) 
    {
      fprintf(stderr,"uniprime: not enough memory\n");
      exit(1);
    }

  if(!(LengthProbe = (int *)  malloc(maxnoprobe * sizeof(int)))) 
    {
      fprintf(stderr,"uniprime: not enough memory\n");
      exit(1);
    }

  for(m=0; m<maxnoprobe; m++)
    {
      if(!(Probe[m]= (char *) malloc(PROBELENGTH * sizeof(char))))
	{
	  fprintf(stderr,"uniprime: not enough memory\n");
	  exit(1);
	}
      if(!(nameProbe[m] =  (char *) malloc(MAXSTRINGP * sizeof(char))))
	{
	  fprintf(stderr,"uniprime: not enough memory\n");
	  exit(1);
	}
    }
  fprintf(stderr,"I have done the double allocation of Probe and nameProbe, and the simple allocation of LengthProbe\n");  

  n=0;
  while (fscanf(file, "%s %s\n", nameProbe[n], Probe[n]) !=EOF)
    {
      //fprintf(stderr,"I am in the probe file scan, n is %d, nameProbe[n] is %s and Probe[n] is %s\n", n, nameProbe[n], Probe[n] );
      n++;
    }
  fprintf(stderr,"I have done the probe file scan and n is %d\n", n); 

  for (m=0;m<n;m++)
    {
      LengthProbe[m] = strlen(Probe[m]);
      //fprintf(stderr,"The length of the probe sequence no %d is %d\n", m, LengthProbe[m]); 
    }
  fprintf(stderr,"I have computed the length of each probe\n"); 



  /**********************************************************************************************************
   read the maximumum number of mismatches tolerated to report a match and start looking for matches 
  ************************************************************************************************************/
  maxnbmismatch = atoi(argv[6]);
  
  shortime();
  //  system("date +%D\" \"%T\" \"%s >&2");
  // align each probe to all genomic sequences 
  m=0;   // probe number 
  while(m<n)
    {
      fprintf(stderr,"Aligning probe # %d to mrna sequences ... ", m+1);
      p=0;   // mrna sequence number 
      while(p<q)
	{
	  i=0;        // position in current mrna sequence 
	  while(i<=LengthGxSeq[p]-LengthProbe[m])   
	    {
	      j=i;          // iterator along the mrna sequence
	      l=0;          // iterator along the probe sequence
	      mismatch=-1;   // current number of mismatches 
	      while((mismatch<maxnbmismatch) && (l<LengthProbe[m]))
		{
		  if (GxSeq[p][j]!=Probe[m][l])
		    mismatch++;
		  j++;
		  l++;
		}  
	      if(mismatch<maxnbmismatch)  // means we have found a match with no more than maxnbmismatch mismatches, so we report it
		fprintf(stdout,"%s\t%s\t%s\t%u\t%d\n", nameProbe[m], Probe[m], nameGxSeq[p], i+1, mismatch+1);

	      i++;
	    }	  
	  p++;
	}
      //      fprintf(stderr,"I have done probe no %d\n", m);
      //      fprintf(stderr,"done; time: %s\n", asctime(localtime(&mytime)));
      fprintf(stderr,"done;\n");
      shortime();
      //      system("date +%D\" \"%T\" \"%s >&2");
      m++;
    }
  fprintf(stderr,"I am finished\n");
}
